#include "window.h"
#include "SDL_rect.h"
#include "SDL_render.h"
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include <SDL_events.h>

void handle_quit(const window_t *wnd, void *obj, const SDL_Event *ev);
void clear_backgroud(const window_t *wnd, void *obj);

bool open_window(window_t *wnd, const char *title, u64 w, u64 h,
                 colour_t bg_colour) {
  SDL_Init(SDL_INIT_EVERYTHING);

  IMG_Init(IMG_INIT_JPG | IMG_INIT_PNG | IMG_INIT_TIF);

  wnd->window =
      SDL_CreateWindow(title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, w,
                       h, SDL_WINDOW_SHOWN);
  if (!wnd->window) {
    close_window(wnd);

    return false;
  }

  wnd->renderer = SDL_CreateRenderer(
      wnd->window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
  if (!wnd->renderer) {
    close_window(wnd);

    return false;
  }

  wnd->width = w;
  wnd->height = h;
  wnd->running = true;
  wnd->bg_colour = bg_colour;
  wnd->event = (SDL_Event){0};
  wnd->count = 0;

  SDL_EventState(SDL_DROPFILE, SDL_ENABLE);

  add_event_listener(wnd, (void *)wnd, handle_quit);
  add_renderer_func(wnd, (void *)wnd, clear_backgroud);

  return true;
}

void run_window_loop(window_t *wnd) {
  while (wnd->running) {
    while (SDL_PollEvent(&wnd->event)) {
      for (u64 i = 0; i < wnd->count; ++i) {
        event_listener_t func = wnd->listeners[i];

        if (!func) {
          continue;
        }

        func(wnd, wnd->objects[i], &wnd->event);
      }
    }

    for (u64 i = 0; i < wnd->count; ++i) {
      render_func_t func = wnd->render_funcs[i];

      if (!func) {
        continue;
      }

      func(wnd, wnd->objects[i]);
    }

    SDL_RenderPresent(wnd->renderer);
  }
}

void close_window(window_t *wnd) {
  wnd->width = 0;
  wnd->height = 0;
  wnd->running = false;
  wnd->bg_colour.abgr = 0;
  wnd->event = (SDL_Event){0};

  if (wnd->renderer) {
    SDL_DestroyRenderer(wnd->renderer);
  }

  if (wnd->window) {
    SDL_DestroyWindow(wnd->window);
  }

  IMG_Quit();

  SDL_Quit();
}

bool add_event_listener(window_t *wnd, void *obj, event_listener_t listener) {
  if (!wnd || wnd->count + 1 >= MAX_HANDLERS_LENGTH) {
    return false;
  }

  u64 index = (wnd->count)++;

  wnd->objects[index] = obj;
  wnd->listeners[index] = listener;

  return true;
}

bool add_renderer_func(window_t *wnd, void *obj, render_func_t func) {
  if (!wnd || wnd->count + 1 >= MAX_HANDLERS_LENGTH) {
    return false;
  }

  u64 index = (wnd->count)++;

  wnd->objects[index] = obj;
  wnd->render_funcs[index] = func;

  return true;
}

void fill_rect(const window_t *wnd, const render_rect_t *rect,
               colour_t colour) {
  SDL_Rect dest = {
      .x = rect->x, .y = rect->y, .w = rect->width, .h = rect->height};

  SDL_SetRenderDrawColor(wnd->renderer, colour.colour.r, colour.colour.g,
                         colour.colour.b, colour.colour.a);

  SDL_RenderFillRect(wnd->renderer, &dest);
}

void draw_rect(const window_t *wnd, const render_rect_t *rect,
               colour_t colour) {
  SDL_Rect dest = {
      .x = rect->x, .y = rect->y, .w = rect->width, .h = rect->height};

  SDL_SetRenderDrawColor(wnd->renderer, colour.colour.r, colour.colour.g,
                         colour.colour.b, colour.colour.a);

  SDL_RenderDrawRect(wnd->renderer, &dest);
}

SDL_Texture *load_texture(const window_t *wnd, const char *path) {
  return IMG_LoadTexture(wnd->renderer, path);
}

void unload_texture(SDL_Texture *texture) { SDL_DestroyTexture(texture); }

void get_texture_size(SDL_Texture *texture, i32 *w, i32 *h) {
  SDL_QueryTexture(texture, NULL, NULL, w, h);
}

void render_texture(const window_t *wnd, SDL_Texture *texture,
                    const render_rect_t *src_rect,
                    const render_rect_t *dst_rect) {
  SDL_Rect src = {
      .x = src_rect ? src_rect->x : 0,
      .y = src_rect ? src_rect->y : 0,
      .w = src_rect ? src_rect->width : 0,
      .h = src_rect ? src_rect->height : 0,
  };
  SDL_Rect dst = {
      .x = dst_rect ? dst_rect->x : 0,
      .y = dst_rect ? dst_rect->y : 0,
      .w = dst_rect ? dst_rect->width : 0,
      .h = dst_rect ? dst_rect->height : 0,
  };

  SDL_RenderCopy(wnd->renderer, texture, src_rect ? &src : NULL, &dst);
}

bool aabb(u64 x, u64 y, const render_rect_t *rect) {
  return x >= rect->x && y >= rect->y && x < rect->x + rect->width &&
         y < rect->y + rect->height;
}

void handle_quit(const window_t *wnd, void *obj, const SDL_Event *ev) {
  if (ev->type == SDL_QUIT) {
    window_t *window = (window_t *)obj;

    window->running = false;
  }
}

void clear_backgroud(const window_t *wnd, void *obj) {
  SDL_SetRenderDrawColor(wnd->renderer, wnd->bg_colour.colour.r,
                         wnd->bg_colour.colour.g, wnd->bg_colour.colour.b,
                         wnd->bg_colour.colour.a);

  SDL_RenderClear(wnd->renderer);
}
